/*
 *   BSD LICENSE
 *
 *   Copyright (C) EZchip Semiconductor Ltd. 2015.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 *     * Neither the name of EZchip Semiconductor nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef _RTE_ATOMIC_LOONGSON_H_
#define _RTE_ATOMIC_LOONGSON_H_

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include "generic/rte_atomic.h"

/**
 * MIPS Synchronize
 */
static inline void synchronize(void)
{
    __asm__ __volatile__("sync":::"memory");
}

/**
 * General memory barrier.
 *
 * Guarantees that the LOAD and STORE operations generated before the
 * barrier occur before the LOAD and STORE operations generated after.
 * This function is architecture dependent.
 */
#define rte_mb() synchronize()

/**
 * Write memory barrier.
 *
 * Guarantees that the STORE operations generated before the barrier
 * occur before the STORE operations generated after.
 * This function is architecture dependent.
 */
#define rte_wmb() synchronize()

/**
 * Read memory barrier.
 *
 * Guarantees that the LOAD operations generated before the barrier
 * occur before the LOAD operations generated after.
 * This function is architecture dependent.
 */
#define rte_rmb() synchronize()

#define rte_smp_mb() rte_mb()

#define rte_smp_wmb() rte_mb()

#define rte_smp_rmb() rte_mb()

#define rte_io_mb() rte_mb()

#define rte_io_wmb() rte_mb()

#define rte_io_rmb() rte_mb()

#define rte_io_mb() rte_mb()

#define rte_cio_wmb() rte_wmb()

#define rte_cio_rmb() rte_rmb()

#if 1 //defined(CONFIG_CPU_LOONGSON3)
#define __LS3A_WAR_LLSC         "       .set mips64r2\nsynci 0\n.set mips0\n"
#define __ls3a_war_llsc()               __asm__ __volatile__("synci 0" : : :"memory")

#else
#define __LS3A_WAR_LLSC
#define __ls3a_war_llsc()
#endif

#define smp_mb__before_llsc()

#define smp_llsc_mb()

/*
 * Helper Functions
 */

static inline void
mips_atomic32_add(rte_atomic32_t *v, int32_t i)
{
    int32_t temp = 0;
    __asm__ __volatile__(
		"1:	ll %[tmp], %[cnt] \n"
		"	add	%[tmp], %[i] \n"
		"	sc %[tmp], %[cnt] \n"
		"	beqzl %[tmp], 1b \n"
		: [tmp] "=&r" (temp), [cnt] "+m" (v->cnt)
		: [i] "Ir" (i));
}

static inline int32_t
mips_atomic32_add_and_return(rte_atomic32_t *v, int32_t i)
{
    int32_t temp = 0;
    int32_t rVal = 0;
    __asm__ __volatile__(
		"1:	ll %[tmp], %[cnt] \n"
		"	add	%[tmp], %[i] \n"
        "   move %[rv], %[tmp] \n"
		"	sc %[tmp], %[cnt] \n"
		"	beqzl %[tmp], 1b \n"
		: [tmp] "=&r" (temp), [cnt] "+m" (v->cnt), [rv] "=&r"(rVal)
		: [i] "Ir" (i));

    return rVal;
}

static inline void
mips_atomic32_sub(rte_atomic32_t *v, int32_t i)
{
    int32_t temp = 0;
    __asm__ __volatile__(
		"1:	ll %[tmp], %[cnt] \n"
		"	sub	%[tmp], %[i] \n"
		"	sc %[tmp], %[cnt] \n"
		"	beqzl %[tmp], 1b \n"
		: [tmp] "=&r" (temp), [cnt] "+m" (v->cnt)
		: [i] "Ir" (i));
}

static inline int32_t
mips_atomic32_sub_and_return(rte_atomic32_t *v, int32_t i)
{
    int32_t temp = 0;
    int32_t rVal = 0;
    __asm__ __volatile__(
		"1:	ll %[tmp], %[cnt] \n"
		"	sub	%[tmp], %[i] \n"
        "   move %[rv], %[tmp] \n"
		"	sc %[tmp], %[cnt] \n"
		"	beqzl %[tmp], 1b \n"
		: [tmp] "=&r" (temp), [cnt] "+m" (v->cnt), [rv] "=&r"(rVal)
		: [i] "Ir" (i));

    return rVal;
}

static inline void
mips_atomic64_add(rte_atomic64_t *v, int64_t i)
{
	int64_t temp = 0;

	__asm__ __volatile__(
		"1:	lld %[tmp], %[cnt] \n"
		"	dadd	%[tmp], %[i] \n"
		"	scd %[tmp], %[cnt] \n"
		"	beqzl %[tmp], 1b \n"
		: [tmp] "=&r" (temp), [cnt] "+m" (v->cnt)
		: [i] "Ir" (i));
}

static inline int64_t
mips_atomic64_add_and_return(rte_atomic64_t *v, int64_t i)
{
	int64_t temp = 0;
    int64_t rVal = 0;

	__asm__ __volatile__(
		"1:	lld %[tmp], %[cnt] \n"
		"	dadd	%[tmp], %[i] \n"
        "   move %[rv], %[tmp] \n"
		"	scd %[tmp], %[cnt] \n"
		"	beqzl %[tmp], 1b \n"
		: [tmp] "=&r" (temp), [cnt] "+m" (v->cnt), [rv] "=&r"(rVal)
		: [i] "Ir" (i));

    return rVal;
}

static inline void
mips_atomic64_sub(rte_atomic64_t *v, int64_t i)
{
	int64_t temp = 0;

	__asm__ __volatile__(
		"1:	lld %[tmp], %[cnt] \n"
		"	dsub %[tmp], %[i] \n"
		"	scd %[tmp], %[cnt] \n"
		"	beqzl %[tmp], 1b \n"
		: [tmp] "=&r" (temp), [cnt] "+m" (v->cnt)
		: [i] "Ir" (i));
}

static inline int64_t
mips_atomic64_sub_and_return(rte_atomic64_t *v, int64_t i)
{
	int64_t temp = 0;
    int64_t rVal = 0;

	__asm__ __volatile__(
		"1:	lld %[tmp], %[cnt] \n"
		"	dsub %[tmp], %[i] \n"
        "   move %[rv], %[tmp] \n"
		"	scd %[tmp], %[cnt] \n"
		"	beqzl %[tmp], 1b \n"
		: [tmp] "=&r" (temp), [cnt] "+m" (v->cnt), [rv] "=&r"(rVal)
		: [i] "Ir" (i));

    return rVal;
}

static inline int
mips_atomic32_cmpset(volatile uint32_t *dst, uint32_t exp, uint32_t src)
{
    uint32_t res = 0, temp = 0;
    __asm__ __volatile__(
        "1: ll %[tmp], %[dst]  \n"
        "   li %[res], 0 \n"
        "   bne %[tmp], %[exp], 2f \n"
        "   move %[tmp], %[src] \n"
        "   sc %[tmp], %[dst] \n"
        "   li %[res], 1 \n"
        "   beqz %[tmp], 1b \n"
        "2: \n"
        : [res] "=&r"(res),
          [tmp] "+&r"(temp),
          [dst] "+m"(*dst)
        : [src] "r"(src),
          [exp] "r"(exp),
          "m"(*dst)
        : "memory"
        );

    return res;
}

static inline int
mips_atomic64_cmpset(volatile uint64_t *dst, uint64_t exp, uint64_t src)
{
    uint32_t res = 0;
    uint64_t temp = 0;
    __asm__ __volatile__(
        "1: lld %[tmp], %[dst]  \n"
        "   li %[res], 0 \n"
        "   bne %[tmp], %[exp], 2f \n"
        "   move %[tmp], %[src] \n"
        "   scd %[tmp], %[dst] \n"
        "   li %[res], 1 \n"
        "   beqz %[tmp], 1b \n"
        "2: \n"
        : [res] "=&r"(res),
          [tmp] "+&r"(temp),
          [dst] "+m"(*dst)
        : [src] "r"(src),
          [exp] "r"(exp),
          "m"(*dst)
        : "memory"
        );

    return res;
}

#ifndef RTE_FORCE_INTRINSICS
static inline int
rte_atomic16_cmpset(volatile uint16_t *dst, uint16_t exp, uint16_t src)
{
	return __sync_bool_compare_and_swap(dst, exp, src);
}

static inline int rte_atomic16_test_and_set(rte_atomic16_t *v)
{
	return rte_atomic16_cmpset((volatile uint16_t *)&v->cnt, 0, 1);
}

static inline void
rte_atomic16_inc(rte_atomic16_t *v)
{
	rte_atomic16_add(v, 1);
}

static inline void
rte_atomic16_dec(rte_atomic16_t *v)
{
	rte_atomic16_sub(v, 1);
}

static inline int rte_atomic16_inc_and_test(rte_atomic16_t *v)
{
	return __sync_add_and_fetch(&v->cnt, 1) == 0;
}

static inline int rte_atomic16_dec_and_test(rte_atomic16_t *v)
{
	return __sync_sub_and_fetch(&v->cnt, 1) == 0;
}

static inline uint16_t
rte_atomic16_exchange(volatile uint16_t *dst, uint16_t val)
{
	return __atomic_exchange_2(dst, val, __ATOMIC_SEQ_CST);
}

/*------------------------- 32 bit atomic operations -------------------------*/

static inline int
rte_atomic32_cmpset(volatile uint32_t *dst, uint32_t exp, uint32_t src)
{
	return mips_atomic32_cmpset(dst, exp, src);
}

static inline int rte_atomic32_test_and_set(rte_atomic32_t *v)
{
	return rte_atomic32_cmpset((volatile uint32_t *)&v->cnt, 0, 1);
}

static inline void
rte_atomic32_inc(rte_atomic32_t *v)
{
	mips_atomic32_add(v, 1); 
}

static inline void
rte_atomic32_dec(rte_atomic32_t *v)
{
	mips_atomic32_sub(v, 1); 
}

static inline int rte_atomic32_inc_and_test(rte_atomic32_t *v)
{
	return mips_atomic32_add_and_return(v, 1) == 0;
}

static inline int rte_atomic32_dec_and_test(rte_atomic32_t *v)
{
	return mips_atomic32_sub_and_return(v, 1) == 0;
}

static inline unsigned long __xchg_u32(volatile uint32_t * m, uint32_t val)
{
	uint32_t retval;

	smp_mb__before_llsc();

	unsigned long dummy;

	__ls3a_war_llsc();
	__asm__ __volatile__(
		"	.set	mips3				\n"
		"	ll	%0, %3		# xchg_u32	\n"
		"	.set	mips0				\n"
		"	move	%2, %z4				\n"
		"	.set	mips3				\n"
		"	sc	%2, %1				\n"
		"	.set	mips0				\n"
		: "=&r" (retval), "=m" (*m), "=&r" (dummy)
		: "R" (*m), "Jr" (val)
		: "memory");

	smp_llsc_mb();

	return retval;
}

static inline uint32_t
rte_atomic32_exchange(volatile uint32_t *dst, uint32_t val)
{
	return __xchg_u32(dst, val);
}
/*------------------------- 64 bit atomic operations -------------------------*/
static inline int
rte_atomic64_cmpset(volatile uint64_t *dst, uint64_t exp, uint64_t src)
{
	return mips_atomic64_cmpset(dst, exp, src);
}

static inline void
rte_atomic64_init(rte_atomic64_t *v)
{
    //int64_t temp = 0;
    //__asm__ __volatile__(
    //    "1: lld %[tmp], %[cnt] \n"
    //    "   li %[tmp], 0 \n"
    //    "   scd %[tmp], %[cnt] \n"
    //    "   beqz %[tmp], 1b \n"
    //    : [cnt] "+m" (v->cnt), [tmp] "+&r" (temp)
    //    : "m"(v->cnt)
    //    : "memory"
    //    );
    v->cnt = 0;
}

static inline int64_t
rte_atomic64_read(rte_atomic64_t *v)
{
	//int64_t temp = 0, res = 0;
    //__asm__ __volatile__(
    //    "1: lld %[tmp], %[cnt] \n"
    //    "   move %[res], %[tmp] \n"
    //    "   scd %[tmp], %[cnt] \n"
    //    "   beqz %[tmp], 1b \n"
    //    : [cnt] "+m" (v->cnt), [res] "=&r" (res), [tmp] "+&r" (temp)
    //    : "m"(v->cnt)
    //    : "memory"
    //    );

    //return res;

    return v->cnt;
}

static inline void
rte_atomic64_set(rte_atomic64_t *v, int64_t new_value)
{
    //uint64_t temp = 0;
    //__asm__ __volatile__(
    //    "1: lld %[tmp], %[cnt] \n"
    //    "   move %[tmp], %[nv] \n"
    //    "   scd %[tmp], %[cnt] \n"
    //    "   beqz %[tmp], 1b \n"
    //    : [cnt] "+m" (v->cnt), [tmp] "+&r" (temp)
    //    : [nv] "r" (new_value), "m"(v->cnt)
    //    : "memory"
    //    );
    v->cnt = new_value;
}

static inline void
rte_atomic64_add(rte_atomic64_t *v, int64_t inc)
{
    mips_atomic64_add(v, inc);
}

static inline void
rte_atomic64_sub(rte_atomic64_t *v, int64_t dec)
{
	mips_atomic64_sub(v, dec);
}

static inline void
rte_atomic64_inc(rte_atomic64_t *v)
{
	mips_atomic64_add(v, 1); 
}

static inline void
rte_atomic64_dec(rte_atomic64_t *v)
{
	mips_atomic64_sub(v, 1); 
}

static inline int64_t
rte_atomic64_add_return(rte_atomic64_t *v, int64_t inc)
{
    return mips_atomic64_add_and_return(v, inc);
}

static inline int64_t
rte_atomic64_sub_return(rte_atomic64_t *v, int64_t dec)
{
    return mips_atomic64_sub_and_return(v, dec);
}

static inline int rte_atomic64_inc_and_test(rte_atomic64_t *v)
{
    return mips_atomic64_add_and_return(v, 1) == 0;
}

static inline int rte_atomic64_dec_and_test(rte_atomic64_t *v)
{
    return mips_atomic64_sub_and_return(v, 1) == 0;
}

static inline int rte_atomic64_test_and_set(rte_atomic64_t *v)
{
	return rte_atomic64_cmpset((volatile uint64_t *)&v->cnt, 0, 1);
}

static inline void rte_atomic64_clear(rte_atomic64_t *v)
{
    rte_atomic64_set(v, 0);
}

static inline uint64_t __xchg_u64(volatile uint64_t * m, uint64_t val)
{
	uint64_t retval;

	smp_mb__before_llsc();

	unsigned long dummy;

	__ls3a_war_llsc();
	__asm__ __volatile__(
		"	.set	mips3				\n"
		"	lld	%0, %3		# xchg_u64	\n"
		"	move	%2, %z4				\n"
		"	scd	%2, %1				\n"
		"	.set	mips0				\n"
		: "=&r" (retval), "=m" (*m), "=&r" (dummy)
		: "R" (*m), "Jr" (val)
		: "memory"
	);

	smp_llsc_mb();

	return retval;
}

static inline uint64_t
rte_atomic64_exchange(volatile uint64_t *dst, uint64_t val)
{
	return __xchg_u64(dst, val);
}
#endif

#ifdef __cplusplus
}
#endif

#endif /* _RTE_ATOMIC_LOONGSON_H_ */
